import ast
from dataclasses import dataclass
import error

from ast import Add, BinOp, Constant, Expression, Mult, Name, Sub, expr, operator
from collections.abc import Mapping

type Value = IntValue | StrValue


@dataclass
class IntValue:
    value: int


@dataclass
class StrValue:
    value: str


def eval_expr(e: expr, m: Mapping[str, Value]) -> Value:
    try:
        match e:
            case Name(id):
                if id in m:
                    return m[id]
                else:
                    raise NameError("name '" + id + "' is not defined")
            case Constant(i) if type(i) is int:
                return IntValue(i)
            case Constant(s) if type(s) is str:
                return StrValue(s)
            case BinOp(left, Add(), right):
                left_value: Value = eval_expr(left, m)
                right_value: Value = eval_expr(right, m)
                match left_value, right_value:
                    case IntValue(i1), IntValue(i2):
                        return IntValue(i1 + i2)
                    case StrValue(s1), StrValue(s2):
                        return StrValue(s1 + s2)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for +")
            case BinOp(left, Sub(), right):
                left_value: Value = eval_expr(left, m)
                right_value: Value = eval_expr(right, m)
                match left_value, right_value:
                    case IntValue(i1), IntValue(i2):
                        return IntValue(i1 - i2)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for -")
            case BinOp(left, Mult(), right):
                left_value: Value = eval_expr(left, m)
                right_value: Value = eval_expr(right, m)
                match left_value, right_value:
                    case IntValue(i1), IntValue(i2):
                        return IntValue(i1 * i2)
                    case IntValue(i1), StrValue(s2):
                        return StrValue(i1 * s2)
                    case StrValue(s1), IntValue(i2):
                        return StrValue(s1 * i2)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for *")

            case _:
                raise ValueError("Unsupported expression")
    except Exception as err:
        error.fail(e, err)


def eval(e: Expression, m: Mapping[str, Value]) -> int | str:
    match eval_expr(e.body, m):
        case IntValue(i):
            return i
        case StrValue(s):
            return s


src = """
(9 - "e") + ("z" * 4)
"""


def main() -> None:
    e: Expression = ast.parse(src, mode='eval')
    print(ast.dump(e, indent=4))
    print(src)
    print(eval(e, {}))


main()
