import ast
import error

from ast import Add, And, BinOp, BoolOp, Call, Compare, Constant, Eq, Expression, Gt, Lt, Mult, Name, Sub, expr, operator
from collections.abc import Mapping
from dataclasses import dataclass

type Value = IntValue | StrValue | BoolValue | FunValue


@dataclass
class FunValue:
    args: list[str]
    body: expr


@dataclass
class BoolValue:
    value: bool


@dataclass
class IntValue:
    value: int


@dataclass
class StrValue:
    value: str


def eval_expr(e: expr, m: Mapping[str, Value]) -> Value:
    try:
        match e:
            case Name(id):
                if id in m:
                    return m[id]
                else:
                    raise NameError("name '" + id + "' is not defined")
            case Constant(b) if type(b) is bool:
                return BoolValue(b)
            case Constant(i) if type(i) is int:
                return IntValue(i)
            case Constant(s) if type(s) is str:
                return StrValue(s)
            case BinOp(left, Add(), right):
                left_value = eval_expr(left, m)
                right_value = eval_expr(right, m)
                match left_value, right_value:
                    case BoolValue(b1), BoolValue(b2):
                        return IntValue(b1 + b2)
                    case BoolValue(b1), IntValue(i2):
                        return IntValue(b1 + i2)
                    case IntValue(i1), BoolValue(b2):
                        return IntValue(i1 + b2)
                    case IntValue(i1), IntValue(i2):
                        return IntValue(i1 + i2)
                    case StrValue(s1), StrValue(s2):
                        return StrValue(s1 + s2)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for +")
            case BinOp(left, Sub(), right):
                left_value = eval_expr(left, m)
                right_value = eval_expr(right, m)
                match left_value, right_value:
                    case BoolValue(b1), BoolValue(b2):
                        return IntValue(b1 - b2)
                    case BoolValue(b1), IntValue(i2):
                        return IntValue(b1 - i2)
                    case IntValue(i1), BoolValue(b2):
                        return IntValue(i1 - b2)
                    case IntValue(i1), IntValue(i2):
                        return IntValue(i1 - i2)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for -")
            case BinOp(left, Mult(), right):
                left_value = eval_expr(left, m)
                right_value = eval_expr(right, m)
                match left_value, right_value:
                    case BoolValue(b1), BoolValue(b2):
                        return IntValue(b1 * b2)
                    case BoolValue(b1), IntValue(i2):
                        return IntValue(b1 * i2)
                    case IntValue(i1), BoolValue(b2):
                        return IntValue(i1 * b2)
                    case IntValue(i1), IntValue(i2):
                        return IntValue(i1 * i2)
                    case BoolValue(b1), StrValue(s2):
                        return StrValue(b1 * s2)
                    case StrValue(s1), BoolValue(b2):
                        return StrValue(s1 * b2)
                    case IntValue(i1), StrValue(s2):
                        return StrValue(i1 * s2)
                    case StrValue(s1), IntValue(i2):
                        return StrValue(s1 * i2)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for *")
            case Compare(left, [Lt()], [right]):
                left_value = eval_expr(left, m)
                right_value = eval_expr(right, m)
                match left_value, right_value:
                    case BoolValue(b1), BoolValue(b2):
                        return BoolValue(b1 < b2)
                    case BoolValue(b1), IntValue(i2):
                        return BoolValue(b1 < i2)
                    case IntValue(i1), BoolValue(b2):
                        return BoolValue(i1 < b2)
                    case IntValue(i1), IntValue(i2):
                        return BoolValue(i1 < i2)
                    case StrValue(s1), StrValue(s2):
                        return BoolValue(s1 < s2)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for *")
            case Compare(left, [Gt()], [right]):
                left_value = eval_expr(left, m)
                right_value = eval_expr(right, m)
                match left_value, right_value:
                    case BoolValue(b1), BoolValue(b2):
                        return BoolValue(b1 > b2)
                    case BoolValue(b1), IntValue(i2):
                        return BoolValue(b1 > i2)
                    case IntValue(i1), BoolValue(b2):
                        return BoolValue(i1 > b2)
                    case IntValue(i1), IntValue(i2):
                        return BoolValue(i1 > i2)
                    case StrValue(s1), StrValue(s2):
                        return BoolValue(s1 > s2)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for *")
            case Compare(left, [Eq()], [right]):
                left_value = eval_expr(left, m)
                right_value = eval_expr(right, m)
                match left_value, right_value:
                    case BoolValue(b1), BoolValue(b2):
                        return BoolValue(b1 == b2)
                    case BoolValue(b1), IntValue(i2):
                        return BoolValue(b1 == i2)
                    case IntValue(i1), BoolValue(b2):
                        return BoolValue(i1 == b2)
                    case IntValue(i1), IntValue(i2):
                        return BoolValue(i1 == i2)
                    case BoolValue(b1), StrValue(s2):
                        return BoolValue(False)
                    case StrValue(s1), BoolValue(b2):
                        return BoolValue(False)
                    case IntValue(i1), StrValue(s2):
                        return BoolValue(False)
                    case StrValue(s1), IntValue(i2):
                        return BoolValue(False)
                    case StrValue(s1), StrValue(s2):
                        return BoolValue(s1 == s2)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for *")
            case BoolOp(And(), [left, right]):
                left_value = eval_expr(left, m)
                right_value = eval_expr(right, m)
                match left_value, right_value:
                    case BoolValue(b1), BoolValue(b2):
                        return BoolValue(b1 and b2)
                    case IntValue(i1), IntValue(i2):
                        return IntValue(i1 and i2)
                    case StrValue(s1), StrValue(s2):
                        return StrValue(s1 and s2)
                    case BoolValue(b1), IntValue(i2):
                        if b1:
                            return IntValue(i2)
                        else:
                            return BoolValue(b1)
                    case IntValue(i1), BoolValue(b2):
                        if bool(i1):
                            return BoolValue(b2)
                        else:
                            return IntValue(i1)
                    case BoolValue(b1), StrValue(s2):
                        if b1:
                            return StrValue(s2)
                        else:
                            return BoolValue(b1)
                    case StrValue(s1), BoolValue(b2):
                        if bool(s1):
                            return BoolValue(b2)
                        else:
                            return StrValue(s1)
                    case IntValue(i1), StrValue(s2):
                        if bool(i1):
                            return StrValue(s2)
                        else:
                            return IntValue(i1)
                    case StrValue(s1), IntValue(i2):
                        if bool(s1):
                            return IntValue(i2)
                        else:
                            return StrValue(s1)
                    case _, _:
                        raise TypeError("unsupported operand type(s) for *")
            case Call(Name(f), args):
                match m.get(f):
                    case None:
                        raise NameError("name '" + f + "' is not defined")
                    case FunValue(params, body):
                        # example: params = ["x", "y"]
                        if len(params) != len(args):
                            raise TypeError("invalid number of argguments")
                        values: list[Value] = [eval_expr(arg, m) for arg in args]
                        # example: values = [IntValue(7), IntValue(2)]
                        bindings: dict[str, Value] = dict(zip(params, values))
                        # example: bindings = {"x": IntValue(7), "y": IntValue(2)}
                        return eval_expr(body, dict(m) | bindings)
                    case _:
                        raise TypeError("name '" + f + "' is not a function")
            case _:
                raise ValueError("Unsupported expression")
    except Exception as err:
        error.fail(e, err)


def eval(e: Expression, m: Mapping[str, Value]) -> int | str | bool:
    match eval_expr(e.body, m):
        case BoolValue(b):
            return b
        case IntValue(i):
            return i
        case StrValue(s):
            return s
        case FunValue(_):
            raise ValueError("Cannot display a function")


src = """
f(z, 3) + 1
"""


def main() -> None:
    e: Expression = ast.parse(src, mode='eval')
    print(ast.dump(e, indent=4))
    print(src)
    f = ast.parse("x + y", mode='eval')
    print(eval(e, {"z": IntValue(7), "f": FunValue(["x", "y"], f.body)}))


main()
